/*
* Fw_GDRL.cpp
* Generator and Day Run Lights Controller firmware
* Created: 16.02.2017
* Author: vdd-pro.ru
*/

#include "Common.h"
#include "Timeouts.h"


// port C
#define ADC_T_PC		0 // temperature sensor
#define ADC_PWR_PC		1 // voltage on PWR input
#define CAL_JMP_PC		2 // calibrate jumper
#define ADC_IGN_PC		7 // voltage on ignition switch input

// port B
#define IGN_SW_PB		0 // ignition switch
#define NEAR_SW_PB		1 // near lamps switch
#define FAR_SW_PB		2 // far lamps switch

// port D
#define LED_PD			0 // led and signal lamp output
#define ADJ_05V_PD		1 // increase output on +0.5V switch
#define ADJ_02V_PD		2 // increase output on +0.2V switch
#define GEN_OUT_PD		3 // OC2B PWM output for alternator
#define FAR_OUT_PD		5 // OC0B PWM output for far lamps
#define NEAR_OUT_PD		6 // OC0A PWM output for near lamps
#define BLOCK_SW_PD		7 // block generator (turn off) switch


void inline TimersInit(){
	// TCNT1 is for a real time clock
	TCCR1A = 0;
	// Normal mode, clk/1024 = 7812.5 Hz (for 8 MHz)
	// 65535 -> ~8.3 sec
	TCCR1B = (1 << CS10) | (0 << CS11) | (1 << CS12);
	// disable timer interrupts
	TIMSK1 = 0;
	
	// TCNT0 - PWM for far and near lamps
	//Compare Output Mode, Fast PWM Mode
	//Clear OC0A on Compare Match, set OC0A at TOP
	//Clear OC0B on Compare Match, set OC0B at TOP
	TCCR0A = (1 << COM0A1) | (1 << COM0B1) | (1 << WGM01) | (1 << WGM00);
	TCCR0B = (1 << CS01); // clk/8 = 1 000 000 Hz
	// disable interrupts
	TIMSK0 = 0;
	
	//TCNT2 - PWM for the alternator
	//Clear OC2B on Compare Match, set OC2B at TOP
	TCCR2A = (1 << COM2B1) | (1 << WGM21) | (1 << WGM20);
	TCCR2B = (1 << CS21); // clk/8 = 1 000 000 Hz
	// disable interrupts
	TIMSK2 = 0;
}


/* ------------------------------------------------------------------------- */

void WDT_off(void){
	cli();
	wdt_reset();
	/* Clear WDRF in MCUSR */
	MCUSR &= ~(1<<WDRF);
	/* Write logical one to WDCE and WDE */
	/* Keep old prescaler setting to prevent unintentional time-out */
	WDTCSR |= (1<<WDCE) | (1<<WDE);
	/* Turn off WDT */
	WDTCSR = 0x00;
	//__enable_interrupt();
}

void WDT_on(void){
	wdt_reset();
	/* Set WDRF in MCUSR */
	MCUSR |= (1<<WDRF);
	/* Start timed sequence */
	WDTCSR |= (1<<WDCE) | (1<<WDE);
	/* Set new prescaler(time-out) value = 1024K (1048576) cycles (8.0 s) */
	WDTCSR = (1<<WDE) | (1<<WDP3) | (1<<WDP0);
	//__enable_interrupt();
}

void inline ADC_SelectChannel(uchar ChN){
	ADMUX = (1<<REFS0)|(1<<REFS1)|(0<<ADLAR) | ChN;// ChN pin as input with internal 1.1 V ref
	ADCSRA |= (1<<ADIF)|(1<<ADSC); // start conversion
}

void inline ADC_Init(void){ // needed for reference configuration
	ADCSRA = (1<<ADEN)|(0<<ADATE)|(1<<ADPS2)|(1<<ADPS1)|(0<<ADPS0); // not-autotrigger mode with clock / 64
	ADCSRB = 0; // free running mode
}



// LED display controller
class cInfoLamp{
	cTimeoutMs toggle_t;
	uchar pwm_cnt;
	uchar pwm_level;
	uchar toggle_count;
	public:
	
	cInfoLamp():toggle_t(300), pwm_level(50), toggle_count(0){
		
	}
	
	void SetBrightness(uchar level){
		pwm_level = level;
		if(pwm_level > 100) pwm_level = 100;
	}
	
	void SetBlinking(uchar blink_cnt){
		uchar tcnt = blink_cnt * 2;
		if(tcnt > toggle_count) toggle_count = tcnt;
	}
	
	
	
	void Handler(){
		if(toggle_count){
			if(toggle_t()){
				toggle_count --;
				Toggle();
			}
		}
		else{
			toggle_t.Reset();
			pwm_cnt ++;
			if(pwm_cnt > 100){
				pwm_cnt = 0;
				if(pwm_level) On();
			}
			if(pwm_cnt == pwm_level){
				Off();
			}
		}
	}
	
	private:
	
	void On(){
		PORTD |= (1 << LED_PD);
	}
	
	void Off(){
		PORTD &= ~(1 << LED_PD);
	}
	
	void Toggle(){
		PORTD ^= (1 << LED_PD);
	}
	
} InfoLamp;


// EEPROM ADC calibration constants
EEMEM uchar eeMagic;
EEMEM uint eeCalPwr, eeCalIgn;
// ADC converter shell
class cAdc{
	uint val_Pwr, val_Ign, val_Temp; // raw values (12-bit)
	public:
	uint U_Pwr_mV, U_Ign_mV; // voltages are in mV
	uint Calib_Pwr, Calib_Ign; // adc values for 12V calibration
	signed char Temp; // degrees C
	
	cAdc(): val_Pwr(0), val_Ign(0), val_Temp(0),
	Calib_Pwr(2792), Calib_Ign(2792){
		// try to load calibration from EEPROM
		uchar magic = eeprom_read_byte(&eeMagic);
		if(magic == 0xAB){ // magic is correct
			Calib_Pwr = eeprom_read_word(&eeCalPwr);
			Calib_Ign = eeprom_read_word(&eeCalIgn);
		}
	}
	
	uint GetBatU() const{
		return U_Ign_mV > U_Pwr_mV ? U_Ign_mV : U_Pwr_mV;
	}
	
	void Handler(){
		//check for ADC conversion finish
		if((ADCSRA & (1<<ADSC)) != 0) // conversion is in progress
		return;
		
		uchar cur_ch = ADMUX & 0x0F;
		switch(cur_ch){
			case ADC_T_PC:
			AdcAvg(val_Temp);
			ADC_SelectChannel(ADC_PWR_PC);
			// convert to degrees
			Temp = ((uint32_t(val_Temp) * 1100 / 4095) // converted to mV
			- 500) // 500 mv is 0 *C
			/ 10; // 10 mV per *C
			break;
			
			case ADC_PWR_PC:
			AdcAvg(val_Pwr);
			ADC_SelectChannel(ADC_IGN_PC);
			if(Calib_Pwr > 0) U_Pwr_mV = 12000 * uint32_t(val_Pwr) / Calib_Pwr;
			break;
			
			case ADC_IGN_PC:
			AdcAvg(val_Ign);
			ADC_SelectChannel(ADC_T_PC);
			if(Calib_Ign > 0) U_Ign_mV = 12000 * uint32_t(val_Ign) / Calib_Ign;
			break;
			
			default:
			ADC_SelectChannel(ADC_T_PC);
		}
	}
	
	void Calibrate_12V(){
		Calib_Ign = val_Ign;
		Calib_Pwr = val_Pwr;
		// store to EEPROM
		eeprom_update_byte(&eeMagic, 0xAB);
		eeprom_update_word(&eeCalPwr, Calib_Pwr);
		eeprom_update_word(&eeCalIgn, Calib_Ign);
	}
	
	private:
	
	void AdcAvg(uint &arg){
		// average ADC value (10 bit) and convert to 12 bit
		arg = ((ADC << 2) + arg * 3) >> 2;
	}
	
} Adc;


// far light switch requires complex logic
class cFarSwitch{
	cTimeoutMs handle_t, _500ms, _1sec;
	uchar click_cnt; // switch clicks counter
	uchar prev_state;
	uchar b_allow_drl;
	uint first_click_t; // timestamp of the first click
	public:
	uchar bAllowDRL; // DRL is allowed
	uchar bFarOn; // far lamps required to be ON
	
	cFarSwitch():handle_t(50), _500ms(500), _1sec(1000), b_allow_drl(1), bAllowDRL(1), bFarOn(0){}
	
	void Handler(){
		if(!handle_t()) return;
		uchar cur_state = (PINB & (1 << FAR_SW_PB)) ? 1 : 0;
		// turn ON far lamps only after 500 msec
		if(cur_state){
			if(!bFarOn && _500ms()){
				bFarOn = 1;
			}
		}
		else{
			if(bFarOn){
				bAllowDRL = 0;
				bFarOn = 0;
				_1sec.Reset();
			}
			_500ms.Reset();
		}
		
		// insert delay before drl on after far light turn off
		if(bAllowDRL != b_allow_drl && _1sec()){
			bAllowDRL = b_allow_drl;
		}
		
		// process DRL turn on|off
		// we must toggle DRL flag after 3 clicks during <3 seconds
		if(cur_state != prev_state){
			if(cur_state){
				// count only ON events
				if(Seconds - first_click_t > 2){ // timeout
					// reset counter
					click_cnt = 1;
					first_click_t = Seconds;
				}
				else{
					click_cnt ++;
					if(click_cnt >= 3){
						// toggle DRL flag
						b_allow_drl = !b_allow_drl;
						// info lamp blinking - many blinks when DRL is ON
						InfoLamp.SetBlinking(b_allow_drl ? 6 : 2);
						//reset counters
						click_cnt = 0;
						first_click_t = Seconds;
					}
					
				}
			}
			prev_state = cur_state;
		}
	}
	
} FarSw;

// ADC calibration jumper
class cCalibSwitch{
	cTimeoutSec apply_t;
	public:
	cCalibSwitch():apply_t(5){}
	void Handler(){
		// apply calibration and store to EEPROM only after 5 second of calibration jumper was shorted
		uchar sw_state = (PINC & (1 << CAL_JMP_PC)) ? 0 : 1; // inverted logic
		if(sw_state){
			if(apply_t()){
				Adc.Calibrate_12V();
				// make led blinking
				InfoLamp.SetBlinking(10);
			}
		}
		else{
			apply_t.Reset();
		}
	}
}CalibSw;

// PWM regulator shell
class cRegulator{
	uint chargeU, // voltage for forced charging
	standbyU;	// voltage for standby
	uint lampsGoodU; // voltage level to freely regulate lamps
	uchar lampsMul; // lamps brightness multiplier, 0...100
	cTimeoutMs reg_t, lamps_decrease_t, drl_show_t, charge_1_sec_t;
	uint genBlockUntilT;
	uchar gen_pwm;
	uchar drl_pwm;
	uchar near_pwm, far_pwm;
	uchar drl_step; // 0 - initial off, 1 - move from 0 to 100%, 2 - move from 100% to 30%
	uint charge_seconds; // charge time back counter
	public:
	cRegulator(): lampsMul(30), reg_t(10), lamps_decrease_t(1000), drl_show_t(20), charge_1_sec_t(1000),
	genBlockUntilT(5), gen_pwm(0), drl_pwm(0), near_pwm(0), far_pwm(0), drl_step(0), charge_seconds(0) {}
	
	void Handler(){
		if(!reg_t()) return;
		
		UpdateTargetU(Adc.Temp);
		
		uint bat_u = Adc.GetBatU();
		
		// update lamps multiplier (0...100%)
		if(bat_u > lampsGoodU){
			// fast increase brightness
			if(lampsMul < 100)	lampsMul++;
		}
		else if (bat_u < lampsGoodU - 500){
			if(bat_u < 10000) lampsMul = 0; // turn lamps off immediately
			// decrease brightness slowly
			if(lampsMul > 0 && lamps_decrease_t()) lampsMul--;
		}
		
		// calculate lamps brightness, maximum allowed output voltage is 13V
		uchar lamps_pwm = uint32_t(lampsMul) * 13000 / (bat_u + 1);
		if(lamps_pwm > 100) lamps_pwm = 100;
		else if(lamps_pwm < 15) lamps_pwm = 0;
		
		// control of the near lamps
		uchar bNearOn = (PINB & (1 << NEAR_SW_PB)) ? 1 : 0;
		if(bNearOn){ // turn on smoothly
			if(near_pwm < lamps_pwm) near_pwm++;
			else if (near_pwm > lamps_pwm) near_pwm --;
		}
		else{
			// turn off near lights
			near_pwm = 0;
		}
		SetNearPwm(near_pwm);
		
		uchar bIgnOn = (PINB & (1 << IGN_SW_PB)) ? 1 : 0;
		// far lamps control
		if(FarSw.bFarOn){ // simply turn ON to the maximum power
			if(far_pwm < lamps_pwm) far_pwm++;
			else if (far_pwm > lamps_pwm) far_pwm --;
			
			drl_step = 0;
		}
		else{ // DRL or turn off
			if(FarSw.bAllowDRL && bIgnOn && !bNearOn){
				// DRL is at 30% of maximum power
				if(drl_step == 0){ // initialize
					drl_pwm = 0;
					drl_step = 1;
				}
				else if(drl_step == 1){ // increase to max
					if(drl_show_t()) drl_pwm ++;
					if(drl_pwm >= 100){
						drl_step = 2;
					}
				}
				else if(drl_step == 2){ /// decrease
					if(drl_show_t()) drl_pwm --;
					if(drl_pwm <= 30){
						drl_step = 3; // finished
					}
				}
				
				far_pwm = lamps_pwm * (uint)drl_pwm / 100;
			}
			else{ // turn OFF
				far_pwm = 0;
				drl_step = 0;
			}
		}
		SetFarPwm(far_pwm);
		
		// process generator control
		if(bIgnOn){
			if(bat_u < 10500 && gen_pwm == 100){ //
				// it looks like alternator is stopped
				genBlockUntilT = Seconds + 5;
			}
			
			if(Seconds < genBlockUntilT || (PIND & (1 << BLOCK_SW_PD)) == 0){ // engine start or gen blocking
				gen_pwm = 0;
				drl_step = 0; // disable DRL when blocked
				if(charge_seconds == 0 && Adc.Temp < 30){
					// force charging during 10 seconds on each degree below 30 *C
					charge_seconds = (30 - Adc.Temp) * 10;
				}
				InfoLamp.SetBlinking(1);
			}
			else{ // work mode
				uint target_u;
				if(charge_seconds > 0){
					target_u = chargeU;
					if(charge_1_sec_t() && bat_u > target_u - 200){
						charge_seconds --; // charging OK
					}
				}
				else{
					target_u = standbyU;
				}
				
				if(bat_u > target_u){
					if(gen_pwm > 0) gen_pwm --;
				}
				else{
					if(gen_pwm < 100) gen_pwm ++;
				}
			}
			
			if(bat_u > 16000){
				gen_pwm = 0; // over voltage
				if(bat_u > 16100){
					InfoLamp.SetBlinking(3);
				}
			}
		}
		else{ // no ignition
			gen_pwm = 0;
		}
		SetGenPwm(gen_pwm);
		
		// set control lamp brightness
		uchar ilamp(0);
		if(bat_u < 12500){ // full brightness is at 10V
			uint du = 12500 - bat_u;
			if(du > 2000) du = 2000;
			ilamp = du / 20;
		}
		InfoLamp.SetBrightness(ilamp);
	}
	
	private:
	void SetNearPwm(uchar pwm){
		if(pwm == 0){
			//set input mode for the pin to totally turn off control
			DDRD &=~(1 << NEAR_OUT_PD);
			PORTD &=~(1 << NEAR_OUT_PD); // force pull-up off
		}
		else{
			if(pwm > 95) pwm = 95;
			DDRD |= (1 << NEAR_OUT_PD); // set as output
		}
		OCR0A = (uint)pwm * 255 / 100;
	}
	
	void SetFarPwm(uchar pwm){
		if(pwm == 0){
			//set input mode for the pin to totally turn off control
			DDRD &=~(1 << FAR_OUT_PD);
			PORTD &=~(1 << FAR_OUT_PD); // force pull-up off
		}
		else{
			if(pwm > 95) pwm = 95;
			DDRD |= (1 << FAR_OUT_PD); // set as output
		}
		OCR0B = (uint)pwm * 255 / 100;
	}
	
	void SetGenPwm(uchar pwm){
		if(pwm == 0){
			//set input mode for the pin to totally turn off control
			DDRD &=~(1 << GEN_OUT_PD);
			PORTD &=~(1 << GEN_OUT_PD); // force pull-up off
		}
		else{
			if(pwm > 95) pwm = 95;
			DDRD |= (1 << GEN_OUT_PD); // set as output
		}
		OCR2B = (uint)pwm * 255 / 100;
	}
	
	// calculate ideal target voltage values
	void UpdateTargetU(signed char Temp){
		int cor_mV = (25 - Temp) * 24; // -24 mV per *C correction for 12 V battery
		if((PIND & (1 << ADJ_05V_PD)) == 0) cor_mV += 500;
		if((PIND & (1 << ADJ_02V_PD)) == 0) cor_mV += 200;
		
		chargeU = 14400 + cor_mV;
		standbyU = 13500 + cor_mV;
		if(chargeU > 15500) chargeU = 15500;
		if(standbyU > 15500) standbyU = 15500;
		
		lampsGoodU = standbyU - 1000;
	};
	
} Reg;

// =====================================================================



int main(void){
	
	WDT_off();
	WDT_on();   // apply WDT timeout
	cli(); // disable interrupts - we don't need any
	
	// ports initial configuration
	DDRB = 0; // all inputs
	PORTB = (1 << 3) | (1 << 4) | (1 << 5); // pull-up all unused pins
	
	
	DDRC = 0; // all inputs
	PORTC = (1 << CAL_JMP_PC) | (1 << 3) | (1 << 4) | (1 << 5); // pull-up
	
	
	DDRD = (1 << LED_PD) | (1 << GEN_OUT_PD) | (1 << FAR_OUT_PD) | (1 << NEAR_OUT_PD);
	PORTD = (1 << ADJ_05V_PD) | (1 << ADJ_02V_PD) | (1 << 4) | (1 << BLOCK_SW_PD); // enable pull-up for inputs
	
	//prepare ADC
	ADC_Init();
	
	// prepare timers
	TimersInit();
	
	// start handling
	InfoLamp.SetBlinking(2);
	cTimeoutMs OneSec(1000);
	
	while(1)
	{
		wdt_reset();
		if(OneSec()){ // 1 second timeout
			Seconds ++;
		}
		
		Adc.Handler();
		Reg.Handler();
		FarSw.Handler();
		CalibSw.Handler();
		InfoLamp.Handler();
	}
}

/* ------------------------------------------------------------------------- */

